/*****************************************************************
 * Plot lines and filled polygons. Input space is database region.
 * Output space and output functions are user defined.
 * Converts input east,north lines and polygons to output x,y
 * and calls user supplied line drawing routines to do the plotting.
 *
 * Handles global wrap-around for lat-lon databases.
 *
 * Does not perform region clipping.
 * Clipping must be done by the line draw routines supplied by the user.
 *
 * Note:
 *  Hopefully, cartographic style projection plotting will be added later.
 *******************************************************************/

#include <stdlib.h>
#include "raster.h"

extern struct Cell_head region;
extern CELL **cell;

double xconv, yconv;
double left, right, top, bottom;
double y_min, y_max;
double nearest(double,double);
int edge (double,double,double,double);
int edge_point(double,int);

#define POINT struct point

POINT
{
    double x;
    int y;
};
 int edge_order(const void *, const void *);
 int row_fill(int,double,double);
 int ifloor(double);
 int iceil(double);


int cell_dot (int x, int y)
{
    if (x >= 0 && x < region.cols && y >= 0 && y < region.rows) {
	cell[y][x] = 1;
    }
    return 0;
}

int move (int x, int y)
{
    cur_x = x;
    cur_y = y;

    return 0;
}

int cont (int x, int y)
{
    if(cur_x < 0 && x < 0) goto set;
    if(cur_y < 0 && y < 0) goto set;
    if(cur_x >= region.cols && x >= region.cols) goto set;
    if(cur_y >= region.rows && y >= region.rows) goto set;

    G_bresenham_line (cur_x, cur_y, x, y, cell_dot);

set:
    move (x, y);

    return 0;
}



 double fabs(double x)
{
    return x>0?x:-x;
}

double G_adjust_east_longitude (double east,double west)
{
    while (east > west + 360.0)
	east -=360.0;
    while (east <= west)
	east += 360.0;
    return east;
}

double G_adjust_easting ( double east, struct Cell_head *region)
{
    if (region->proj == PROJECTION_LL)
    {
		east = G_adjust_east_longitude(east, region->west);
		if (east > region->east && east == region->west + 360)
			east = region->west;
    }
    return east;
}

/*
 * G_setup_plot (t, b, l, r, thismove, thiscont)
 *     double t, b, l, r;
 *     int (*thismove)(), (*thiscont)();
 *
 * initialize the plotting capability.
 *    t,b,l,r:   top, bottom, left, right of the output x,y coordinate space.
 *    thismove,thiscont: subroutines that will draw lines in x,y space.
 *       thismove(x,y)   thismove to x,y (no draw)
 *       thiscont(x,y)   draw from previous position to x,y
 * Notes:
 *   thiscont() is responsible for clipping.
 *   The t,b,l,r are only used to compute coordinate transformations.
 *   The input space is assumed to be the current GRASS region.
 */
int G_setup_plot (double t,double b,double l,double r)
{
    left = l;
    right = r;
    top = t;
    bottom = b;

    xconv = (right-left)/(region.east-region.west);
    yconv = (bottom-top)/(region.north-region.south);

    if (top < bottom)
    {
		y_min = iceil(top);
		y_max = ifloor(bottom);
    }
    else
    {
		y_min = iceil(bottom);
		y_max = ifloor(top);
    }
    return 0;
}

#define X(e) (left + xconv * ((e) - region.west))
#define Y(n) (top + yconv * (region.north - (n)))

#define EAST(x) (region.west + ((x)-left)/xconv)
#define NORTH(y) (region.north - ((y)-top)/yconv)

int G_plot_where_xy (double east, double north, int *x, int *y)
{
    *x = ifloor(X(G_adjust_easting(east,&region))+0.5);
    *y = ifloor(Y(north)+0.5);
    return 0;
}

int G_plot_where_en (int x, int y, double *east, double *north)
{
    *east = G_adjust_easting(EAST(x),&region);
    *north = NORTH(y);

    return 0;
}

int G_plot_point (double east, double north)
{
    int x,y;

    G_plot_where_xy(east,north,&x,&y);
    move (x,y);
    cont (x,y);
    return 0;
}

/*
 * G_plot_polygon (x, y, n)
 *
 *    double *x       x coordinates of vertices
 *    double *y       y coordinates of vertices
 *    int n           number of verticies
 *
 * polygon fill from map coordinate space to plot x,y space.
 *     for lat-lon, handles global wrap-around as well as polar polygons.
 *
 * returns 0 ok, 2 n<3, -1 weird internal error, 1 no memory
 */

 POINT *P;
 int np;
 int npalloc = 0;

#define OK 0
#define TOO_FEW_EDGES 2
#define NO_MEMORY 1
#define OUT_OF_SYNC -1

 double nearest(double e0,double e1)
{
    while (e0 - e1 > 180)
	e1 += 360.0;
    while (e1 - e0 > 180)
	e1 -= 360.0;

    return e1;
}

int G_plot_polygon (double *x,double *y,int n)
{
    int i;
    int pole;
    double x0,x1;
    double y0,y1;
    double shift,E,W=0L;
    double e0,e1;
    int shift1, shift2;

    if (n < 3)
        return TOO_FEW_EDGES;

/* traverse the perimeter */

    np = 0;
    shift1 = 0;

/* global wrap-around for lat-lon, part1 */
    if (region.proj == PROJECTION_LL)
    {
	/*
	pole = G_pole_in_polygon(x,y,n);
	*/
pole = 0;

	e0 = x[n-1];
	E = W = e0;

	x0 = X(e0);
	y0 = Y(y[n-1]);

	if (pole &&!edge (x0, y0, x0, Y(90.0*pole)))
		return NO_MEMORY;

	for (i = 0; i < n; i++)
	{
	    e1 = nearest (e0, x[i]);
	    if (e1 > E) E = e1;
	    if (e1 < W) W = e1;

	    x1 = X(e1);
	    y1 = Y(y[i]);

	    if(!edge (x0, y0, x1, y1))
		return NO_MEMORY;

	    x0 = x1;
	    y0 = y1;
	    e0 = e1;
	}
	if (pole &&!edge (x0, y0, x0, Y(90.0*pole)))
		return NO_MEMORY;

	shift = 0;        /* shift into region */
	while (E+shift > region.east)
	    shift -= 360.0;
	while (E+shift < region.west)
	    shift += 360.0;
	shift1 = (int)(X(x[n-1]+shift) - X(x[n-1]));
    }
    else
    {
	x0 = X(x[n-1]);
	y0 = Y(y[n-1]);

	for (i = 0; i < n; i++)
	{
	    x1 = X(x[i]);
	    y1 = Y(y[i]);
	    if(!edge (x0, y0, x1, y1))
		return NO_MEMORY;
	    x0 = x1;
	    y0 = y1;
	}
    }

/* check if perimeter has odd number of points */
    if (np%2)
        return OUT_OF_SYNC;

/* sort the edge points by col(x) and then by row(y) */
    qsort (P, np, sizeof(POINT), &edge_order);

/* plot */
    for (i = 1; i < np; i += 2)
    {
        if (P[i].y != P[i-1].y)
	    return OUT_OF_SYNC;
		row_fill (P[i].y, P[i-1].x+shift1, P[i].x+shift1);
    }
    if (region.proj == PROJECTION_LL)	/* now do wrap-around, part 2 */
    {
	shift = 0;
	while (W+shift < region.west)
	    shift += 360.0;
	while (W+shift > region.east)
	    shift -= 360.0;
	shift2 = (int)(X(x[n-1]+shift) - X(x[n-1]));
	if (shift2 != shift1)
	{
	    for (i = 1; i < np; i += 2)
	    {
		row_fill (P[i].y, P[i-1].x+shift2, P[i].x+shift2);
	    }
	}
    }
    return OK;
}

 int edge (double x0,double y0,double x1,double y1)
{
    register double m;
    double dy, x;
    int ystart, ystop;


/* tolerance to avoid FPE */
    dy = y0 - y1;
    if (fabs(dy) < 1e-10)
	return 1;

    m = (x0 - x1) / dy;

    if (y0 < y1)
    {
        ystart = iceil  (y0);
	ystop  = ifloor (y1);
	if (ystop == y1) ystop--; /* if line stops at row center, don't include point */
    }
    else
    {
        ystart = iceil  (y1);
	ystop  = ifloor (y0);
	if (ystop == y0) ystop--; /* if line stops at row center, don't include point */
    }
    if (ystart > ystop)
	return 1;	/* does not cross center line of row */

    x = m * (ystart - y0) + x0;
    while (ystart <= ystop)
    {
	if(!edge_point (x, ystart++))
	    return 0;
	x += m;
    }
    return 1;
}

 int edge_point( double x, register int y)
{

    if (y < y_min || y > y_max)
	return 1;
    if (np >= npalloc)
    {
	if (npalloc > 0)
	{
	    npalloc *= 2;
	    P = (POINT *) realloc (P, npalloc * sizeof (POINT));
	}
	else
	{
	    npalloc = 32;
	    P = (POINT *) malloc (npalloc * sizeof (POINT));
	}
	if (P == NULL)
	{
	    npalloc = 0;
	    return 0;
	}
    }
    P[np].x   = x;
    P[np++].y = y;
    return 1;
}

 int edge_order(const void *aa, const void *bb)
{
    const struct point *a = (const struct point *)aa;
	const struct point *b = (const struct point *)bb;
    if (a->y < b->y) return (-1);
    if (a->y > b->y) return (1);

    if (a->x < b->x) return (-1);
    if (a->x > b->x) return (1);

    return (0);
}

 int row_fill(int y,double x1,double x2)
{
    int i1,i2;

    i1 = iceil(x1);
    i2 = ifloor(x2);
    if (i1 <= i2)
    {
		move (i1, y);
		cont (i2, y);
	}
    return 0;
}

 int ifloor(double x)
{
    int i;
    i = (int) x;
    if (i > x) i--;
    return i;
}

 int iceil(double x)
{
    int i;

    i = (int) x;
    if (i < x) i++;
    return i;
}

