/* JUnit tests for CollectionTypes class.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */
 
package org.nddp;

import org.nddp.CollectionTypes;

import ptolemy.data.IntToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NamedObj;

import junit.framework.TestCase;

/** 
 @author Timothy M. McPhillips
 */

public class TestCollectionTypes  extends TestCase {

	///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 
    
    public void setUp() {
        	CollectionTypes.clear();
    }
    
    public void testClear() throws Exception {
        
		CollectionTypes.register("C", ACollectionSubclass.class);
        	CollectionTypes.clear();

        	try {
   	        	CollectionTypes.valueOfString("C");

   	        	fail("An IllegalActionException should be thrown because _clear() " +
   	        	        "was called after adding ClassC.");
        	} 
        catch(IllegalActionException expected) {
            assertEquals("No collection type C", expected.getMessage());
        }
    }
    
    public void testIsDefined_EmptyTypeName() throws Exception {

		assertEquals(false, CollectionTypes.isDefined(""));
    }
    
    public void testIsDefined_Match() throws Exception {

        	CollectionTypes.register("C", ACollectionSubclass.class);
		assertEquals(true, CollectionTypes.isDefined("C"));
    }
    
    public void testIsDefined_NoMatch() throws Exception {

		assertEquals(false, CollectionTypes.isDefined("C"));
    }
    
    public void testIsDefined_NullTypeName() throws Exception {

        try {
			CollectionTypes.isDefined(null);
			
			fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
    }
    public void testRegister() throws Exception {

		CollectionTypes.register("C", ACollectionSubclass.class);
		assertEquals(ACollectionSubclass.class, CollectionTypes.valueOfString("C"));
	}
    
    public void testRegister_NoMatchClass() throws Exception {

        try  {
			CollectionTypes.register("B", CollectionTypes.NoMatch.class);
			
			fail("An AssertionError should be thrown because NoMatch " +
		        "was passed for typeClass");
        }
        catch(AssertionError expected) {
            assertEquals("Class NoMatch not allowed.", 
            		expected.getMessage());
        }
	}
    
    public void testRegister_NotCollectionSubclass() throws Exception {

        try  {
			CollectionTypes.register("E", NotACollectionSubclass.class);
			
			fail("An AssertionError should be thrown because a " +
		        "class not subclassed from Collection " +
		        "was passed for typeClass");
        }
        catch(AssertionError expected) {
            assertEquals("class org.nddp.TestCollectionTypes$" +
                	"NotACollectionSubclass " +
                	"must be a subclass of Collection.", 
        			expected.getMessage());
        }
	}
    
    public void testRegister_NullTypeClass() throws Exception {

        try  {
			CollectionTypes.register("C", null);
			
			fail("An AssertionError should be thrown because null " +
		        "was passed for typeClass");
        }
        catch(AssertionError expected) {
            assertEquals("Null typeClass not allowed.", 
            		expected.getMessage());
        }
	}
    
    public void testRegister_NullTypeName() throws Exception {

        try  {
			CollectionTypes.register(null, ACollectionSubclass.class);
			
			fail("An AssertionError should be thrown because null" +
		        "was passed for typeName");
        }
        catch(AssertionError expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
	}
    
    public void testRegisterDefaultCollectionTypes() throws Exception {
    
        CollectionTypes.registerDefaultCollectionTypes();
        
        assertEquals(ProjectCollection.class, 
       		CollectionTypes.valueOfString("Project"));

        assertEquals(TextFileCollection.class, 
       		CollectionTypes.valueOfString("TextFile"));        
    }

    public void testValueOfParameter_EmptyClassName() throws Exception {
        
        	CollectionTypes.registerDefaultCollectionTypes();
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken(""));
        	assertEquals(Object.class, CollectionTypes.valueOfParameter(parameter));
    } 
    
    public void testValueOfParameter_Match() throws Exception {
        
		CollectionTypes.register("C", ACollectionSubclass.class);
		
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken("C"));
        	
        	assertEquals(ACollectionSubclass.class, 
    	        	CollectionTypes.valueOfParameter(parameter));
    }
    
    public void testValueOfParameter_NoMatch() throws Exception {
        
        	Parameter parameter = new Parameter(
			new NamedObj(), "parameter", new StringToken("C"));
        	
        	try { 
        	    CollectionTypes.valueOfParameter(parameter);
	        
	        fail("An IllegalActionException should be thrown" +
	                "because Class C has not been added.");
        	}
         catch(IllegalActionException expected) {
            assertEquals("No collection type C", expected.getMessage());
        }
    }
    
    public void testValueOfParameter_NullParameter() throws Exception {
        
        	try {
	        CollectionTypes.valueOfParameter(null);
	        
	        fail("A NullPointerException should be thrown" +
	        		"because null was passed for the parameter.");
        	}
        catch(NullPointerException expected) {
            assertEquals("Null collectionTypeParameter not allowed.", 
            		expected.getMessage());
        }
    }

    public void testValueOfString_EmptyClassName() throws Exception {
        
        CollectionTypes.registerDefaultCollectionTypes();
        assertEquals(Object.class, CollectionTypes.valueOfString(""));
    }
    
    public void testValueOfString_Match() throws Exception {
        
		CollectionTypes.register("C", ACollectionSubclass.class);
        	assertEquals(ACollectionSubclass.class, 
        	        CollectionTypes.valueOfString("C"));
    }
    
    public void testValueOfString_NoMatch() throws Exception {
        
        try {
	        CollectionTypes.valueOfString("C");
	        
	        fail("An IllegalActionException should be thrown" +
	                "because class C has not been added.");
        }
        catch(IllegalActionException expected) {
            assertEquals("No collection type C", expected.getMessage());
        }
    }

    public void testValueOfString_NullTypeName() throws Exception {
        
        	try {
	        CollectionTypes.valueOfString(null);
	        
	        fail("A NullPointerException should have been thrown.");
        	}
        catch(NullPointerException expected) {
            assertEquals("Null typeName not allowed.", 
            		expected.getMessage());
        }
    }

    public void testValueOfToken_EmptyClassName() throws Exception {
        
        	CollectionTypes.registerDefaultCollectionTypes();
        	Token token = new StringToken("");
        	assertEquals(Object.class, CollectionTypes.valueOfToken(token));
    } 
    
    public void testValueOfToken_Match() throws Exception {
        
		CollectionTypes.register("C", ACollectionSubclass.class);
		
        	Token token = new StringToken("C");
        	
        	assertEquals(ACollectionSubclass.class, 
    	        	CollectionTypes.valueOfToken(token));
    }
    
    public void testValueOfToken_NoMatch() throws Exception {
        
        	Token token = new StringToken("C");
        	
        	try { 
        	    CollectionTypes.valueOfToken(token);
	        
	        fail("An IllegalActionException should be thrown" +
	                "because Class C has not been added.");
        	}
         catch(IllegalActionException expected) {
            assertEquals("No collection type C", expected.getMessage());
        }
    }
    
    public void testValueOfToken_NullToken() throws Exception {
        
        	try {
	        CollectionTypes.valueOfToken(null);
	        
	        fail("A NullPointerException should be thrown" +
	        		"because null was passed for the token.");
        	}
        catch(NullPointerException expected) {
            assertEquals("Null collectionTypeToken not allowed.", 
            		expected.getMessage());
        }
    }
    
    public void testValueOfToken_IntToken() throws Exception {
        
        	Token token = new IntToken(0);
        	
        	try { 
        	    CollectionTypes.valueOfToken(token);
	        
	        fail("An ClassCastExceptoin should be thrown" +
	                "because passed token is not a StringToken");
        }
         catch(IllegalActionException expected) {
            assertEquals("collectionTypeToken must " +
	    	        "be a StringToken", expected.getMessage());
        }
    }

    ///////////////////////////////////////////////////////////////////
	////                      inner classes                        ////   
    
	class ACollectionSubclass extends Collection {
	
	    public String type() {
	        return "C";
	    }
	}

	class NotACollectionSubclass {}
}


