// This class will process the climate data in IPCC climate layers
// into a text format that can be manipulated by GIS software
// The original climate data is stored with monthly data all in
// a single file. This class calculates annual and three month
// averages and seperates the data into one file per month
//
// based on code originally written by Steven Philips
// phillips@research.att.com

// modified by Dan Higgins(DEFH) (higgins@nceas.ucsb.edu) Oct 19, 2004

import java.io.*;
import java.util.*;

public class Process {
    static int months=12, rows=360, cols=720, NODATA_value=-9999;

    static short[][][] readFile (String fn) throws IOException {
	    BufferedReader in = new BufferedReader(new FileReader(fn));
	    in.readLine(); in.readLine();
	    short[][][] result = new short[months][rows][cols];
      short temp;
	    for (int m=0; m<months; m++)
	    for (int r=0; r<rows; r++) {
		    String line = in.readLine();
		    for (int c=0; c<cols; c++) {
		      result[m][r][c] = Short.parseShort(line.substring(c*5, (c+1)*5).trim());
        } 
        // swap column data so that England is in center - DFH
        for (int c1=0; c1<cols/2; c1++) {
          temp = result[m][r][c1];
          result[m][r][c1] = result[m][r][c1+cols/2];  
          result[m][r][c1+cols/2] = temp;
        }
	    }
	return result;
    }

    public static PrintWriter header(String fn) {
	PrintWriter out=null;
	try {
	    out = new PrintWriter(new FileOutputStream(fn));
	}
	catch (FileNotFoundException e) {
	    System.out.println("Problem opening output file " + fn + ": " + e.toString());
	    System.exit(1);
	}
	out.println("ncols         720");
	out.println("nrows         360");
	out.println("xllcorner     -180.0");
	//out.println("xllcorner     0.0");
	//	out.println("xllcorner     0.25");
	out.println("yllcorner     -90");
	out.println("cellsize      0.5");
	out.println("NODATA_value  -9999");
	return out;
    }

    static boolean contains(String[] args, String item) {
	for (int i=0; i<args.length; i++)
	    if (args[i].equals(item)) return true;
	return false;
    }

    public static void main(String[] args) {
	String[] vars = {"tmn", "tmx", "cld", "dtr", "frs", "pre", "rad", "tmp", "vap", "wet", "wnd"};
	char[] annType = {'m', 'M', 'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a'};
	//	char[] annType = {'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a', 'a'};
	int v, m, r, c;
	for (v=0; v<vars.length; v++) {
	    short[][][] data = null;
	    System.out.println("Starting " + vars[v]);
	    try {
		data = readFile("c" + vars[v] + "6190.dat");
	    } catch (IOException e) {}
	    int[][] annual = new int[rows][cols];
	    int[][] winter = new int[rows][cols];    // dec-feb
	    int[][] spring = new int[rows][cols];    // mar-may
	    int[][] summer = new int[rows][cols];    //jun-aug
	    int[][] fall   = new int[rows][cols];    //sep-nov
	    boolean[][] missingData = new boolean[rows][cols];

	    for (m=0; m<months; m++) {
		String fileName = vars[v] + "6190_l" + (m+1) + ".asc";
		if (args.length>0 && !contains(args, fileName)) continue;
		PrintWriter out = header("processed/" + fileName);
		for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++)
			out.print(data[m][r][c] + " ");
		    out.println();
		}
		out.close();
	    }
	    for (m=0; m<months; m++) {
		    for (r=0; r<rows; r++) {
		      for (c=0; c<cols; c++) {
			      if (data[m][r][c]==NODATA_value) 
			        missingData[r][c] = true;
			      switch(annType[v]) {
			        case 'm': if (m==0 | data[m][r][c] < annual[r][c]) annual[r][c]=data[m][r][c];
			        break;
			        case 'M': if (m==0 | data[m][r][c] > annual[r][c]) annual[r][c]=data[m][r][c];
			        break;
			        default: annual[r][c] += data[m][r][c];
              break;                
			      }
		      }
		    }
	    }

      // spring - mar-may
	    for (m=3; m<6; m++) {
		    for (r=0; r<rows; r++) {
		      for (c=0; c<cols; c++) {
			      if (data[m][r][c]==NODATA_value) 
			        missingData[r][c] = true;
			      switch(annType[v]) {
			        case 'm': if (m==3 | data[m][r][c] < spring[r][c]) spring[r][c]=data[m][r][c];
			        break;
			        case 'M': if (m==3 | data[m][r][c] > spring[r][c]) spring[r][c]=data[m][r][c];
			        break;
			        default: spring[r][c] += data[m][r][c];
              break;                
			      }
		      }
		    }
	    }

      // summer - jun-aug
	    for (m=6; m<9; m++) {
		    for (r=0; r<rows; r++) {
		      for (c=0; c<cols; c++) {
			      if (data[m][r][c]==NODATA_value) 
			        missingData[r][c] = true;
			      switch(annType[v]) {
			        case 'm': if (m==6 | data[m][r][c] < summer[r][c]) summer[r][c]=data[m][r][c];
			        break;
			        case 'M': if (m==6 | data[m][r][c] > summer[r][c]) summer[r][c]=data[m][r][c];
			        break;
			        default: summer[r][c] += data[m][r][c];
              break;                
			      }
		      }
		    }
	    }

      // fall - sep-nov
	    for (m=9; m<11; m++) {
		    for (r=0; r<rows; r++) {
		      for (c=0; c<cols; c++) {
			      if (data[m][r][c]==NODATA_value) 
			        missingData[r][c] = true;
			      switch(annType[v]) {
			        case 'm': if (m==9 | data[m][r][c] < fall[r][c]) fall[r][c]=data[m][r][c];
			        break;
			        case 'M': if (m==9 | data[m][r][c] > fall[r][c]) fall[r][c]=data[m][r][c];
			        break;
			        default: fall[r][c] += data[m][r][c];
              break;                
			      }
		      }
		    }
	    }

      // winter - dec-feb
      for (r=0; r<rows; r++) {
        for (c=0; c<cols; c++) {
          winter[r][c]=data[11][r][c];      // start with values for dec 
        }
      }
	    for (m=0; m<2; m++) {
		    for (r=0; r<rows; r++) {
		      for (c=0; c<cols; c++) {
			      if (data[m][r][c]==NODATA_value) 
			        missingData[r][c] = true;
			      switch(annType[v]) {
			        case 'm': if (data[m][r][c] < winter[r][c]) winter[r][c]=data[m][r][c];
			        break;
			        case 'M': if (data[m][r][c] > winter[r][c]) winter[r][c]=data[m][r][c];
			        break;
			        default: winter[r][c] += data[m][r][c];
              break;                
			      }
		      }
		    }
	    }


	    String fileName = vars[v] + "6190_ann.asc";
	    if (args.length>0 && !contains(args, fileName)) continue;
	    PrintWriter out = header("processed/" + fileName);
	    for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++) {
		      if (annType[v]=='a')
			      annual[r][c] = (int)Math.round(annual[r][c]/3.0);
		      if (missingData[r][c])
			      annual[r][c] = NODATA_value;
		      out.print(annual[r][c] + " ");
		    }
		    out.println();
	    }
	    out.close(); 

	    fileName = vars[v] + "6190_win.asc";
	    if (args.length>0 && !contains(args, fileName)) continue;
	    out = header("processed/" + fileName);
	    for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++) {
		      if (annType[v]=='a')
			      winter[r][c] = (int)Math.round(winter[r][c]/3.0);
		      if (missingData[r][c])
			      winter[r][c] = NODATA_value;
		      out.print(winter[r][c] + " ");
		    }
		    out.println();
	    }
	    out.close();

	    fileName = vars[v] + "6190_spr.asc";
	    if (args.length>0 && !contains(args, fileName)) continue;
	    out = header("processed/" + fileName);
	    for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++) {
		      if (annType[v]=='a')
			      spring[r][c] = (int)Math.round(spring[r][c]/3.0);
		      if (missingData[r][c])
			      spring[r][c] = NODATA_value;
		      out.print(winter[r][c] + " ");
		    }
		    out.println();
	    }
	    out.close();

	    fileName = vars[v] + "6190_sum.asc";
	    if (args.length>0 && !contains(args, fileName)) continue;
	    out = header("processed/" + fileName);
	    for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++) {
		      if (annType[v]=='a')
			      summer[r][c] = (int)Math.round(summer[r][c]/3.0);
		      if (missingData[r][c])
			      summer[r][c] = NODATA_value;
		      out.print(winter[r][c] + " ");
		    }
		    out.println();
	    }
	    out.close();

	    fileName = vars[v] + "6190_aut.asc";
	    if (args.length>0 && !contains(args, fileName)) continue;
	    out = header("processed/" + fileName);
	    for (r=0; r<rows; r++) {
		    for (c=0; c<cols; c++) {
		      if (annType[v]=='a')
			      fall[r][c] = (int)Math.round(fall[r][c]/12.0);
		      if (missingData[r][c])
			      fall[r][c] = NODATA_value;
		      out.print(winter[r][c] + " ");
		    }
		    out.println();
	    }
	    out.close();

	  }
  }
}
