/**
 *    '$RCSfile: DelimitedParser.java,v $'
 *
 *     '$Author: ruland $'
 *       '$Date: 2005/11/01 20:39:07 $'
 *   '$Revision: 1.2 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.util;


import java.util.Vector;

import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Transformer;
import ptolemy.actor.parameters.PortParameter;
import ptolemy.data.ArrayToken;
import ptolemy.data.IntToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;


//////////////////////////////////////////////////////////////////////////
//// DelimitedParser
/**
   Parse an input string into an array of array of strings that
   represent the rows and columns of a delimited string table.
   It is assumed that each row of the ascii table ends with a '\n'
   (i.e each row is on a single line) and columns are separated by
   a 'delimiter' supplied on a port. All data in a column is of the
   same type, but some columns may represent numbers, others strings,
   etc. so all data is kept in string format here. Each row is a
   string array; an array of arrays represents the whole table. 
   @author Dan Higgins
   @version $Id: DelimitedParser.java,v 1.2 2005/11/01 20:39:07 ruland Exp $
*/

public class DelimitedParser extends Transformer {

    /** Construct an actor with the given container and name.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the actor cannot be contained
     *   by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *   actor with this name.
     */
    public DelimitedParser(CompositeEntity container, String name)
            throws NameDuplicationException, IllegalActionException  {
        super(container, name);

        // Create new parameters and ports.
        // Set default values of the parameters and type constraints.

        outputFormat = new Parameter(this, "outputFormat");
        outputFormat.setStringMode(true);
        outputFormat.setExpression("row");
        outputFormat. addChoice("column");
        outputFormat. addChoice("row"); 
        _outputFormat = _ROW;


        delimiter = new PortParameter(this, "delimiter");
        delimiter.setExpression("\",\"");
        delimiter.setTypeEquals(BaseType.STRING);
//        new Attribute(delimiter.getPort(), "_showName");

        numberColumns = new PortParameter(this, "numberColumns");
        numberColumns.setExpression("1");
        numberColumns.setTypeEquals(BaseType.INT);
//        new Attribute(numberColumns.getPort(), "_showName");   

        numberHeaderLines = new PortParameter(this, "numberHeaderLines");
        numberHeaderLines.setExpression("0");
        numberHeaderLines.setTypeEquals(BaseType.INT);
        
        stringOutput = new TypedIOPort(this, "stringOutput", false, true);
//        stringOutput.setTypeEquals(new ArrayType(BaseType.STRING));
        stringOutput.setTypeEquals(BaseType.GENERAL);

        input.setTypeEquals(BaseType.STRING);
//        output.setTypeEquals(new ArrayType(new ArrayType(BaseType.STRING)));
        output.setTypeEquals(BaseType.GENERAL);
    }

    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////

    /** Parameter that stores a string indicating the format of the output
     *  array. 'column' indicates an array of columns; 'row' indicates an
     *  array of rows
     */
    public Parameter outputFormat;

    /** The string used to delimit colums,
     *  include '\t'.
     */
    public PortParameter delimiter;

    /** The number of columns to be assumed. This is used for cases
     *  where there may be insufficient data or too much data. If missing
     *  the number of tokens in the first row is used.
     */
    public PortParameter numberColumns;

    /** The number of headerLines to be assumed. These lines are stripped.
     */
    public PortParameter numberHeaderLines;

    /**
     *  The output as a single array of strings separated by ','s
     */
    public TypedIOPort stringOutput;

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Determine the output format 
     *  @param attribute The attribute that changed.
     *  @exception IllegalActionException If the function is not recognized.
     */
    public void attributeChanged(Attribute attribute)
            throws IllegalActionException {
        if (attribute == outputFormat) {
            // Use getToken() rather than getExpression()
            // so substitutions occur.
            String outputFormatName
                = ((StringToken)outputFormat.getToken()).stringValue();

            if (outputFormatName.equals("row")) {
                _outputFormat = _ROW;
            } else if (outputFormatName.equals("column")) {
                _outputFormat = _COLUMN;
            } else {
                throw new IllegalActionException(this,
                        "Unrecognized outpurFormat: " + outputFormatName);
            }
        } else {
            super.attributeChanged(attribute);
        }
    }

    /** If there is an input string, parse the string into a number of rows
     *  indicated by newline characters. If no newlines, a single row is 
     *  assumed. Each row is then parsed into columns indicated by the
     *  delimiter string. The result is an array of rows, each of which is
     *  is an array of column strings.
     */
    public void fire() throws IllegalActionException {
        super.fire();
        delimiter.update();
        numberColumns.update();
        if (input.hasToken(0)) {
            StringToken inputToken = (StringToken)input.get(0);
            String value = inputToken.stringValue();
            String delimiterValue = ((StringToken)delimiter.getToken()).stringValue();
            int numberColumnsValue = ((IntToken)numberColumns.getToken()).intValue();
            int numHeaderLines = ((IntToken)numberHeaderLines.getToken()).intValue(); 
            int numRows = 1;
            
            DelimitedReader dr = new DelimitedReader(value,
                            numberColumnsValue, delimiterValue,
                            numHeaderLines, "\n",
                            numRows);

            Vector[] datavec = dr.getTokenizedData(true);  

            if (_outputFormat == _ROW) {
                Token[] table = new Token[datavec.length];
                Token[] strtable = new Token[datavec.length];
                for (int i = 0; i < datavec.length; i++) {//step through each row 
                  String totstr = "";
                  Token[] rows = new Token[datavec[i].size()];
                  for (int j = 0; j < datavec[i].size(); j++) {
                    String temp = (String)datavec[i].elementAt(j);
                    StringToken str = new StringToken(temp);
                    rows[j] = str;
                    totstr = totstr + temp + ",";
                  }
                  totstr = totstr.substring(0, totstr.length()-1);
                  StringToken strrow = new StringToken(totstr);
                  strtable[i] = strrow;
                  table[i] = new ArrayToken(rows);
                } 
            output.send(0, new ArrayToken(table));
            stringOutput.send(0, new ArrayToken(strtable));
             }


            if (_outputFormat == _COLUMN) {
                Token[] coltable = new Token[numberColumnsValue];
                Token[] strcoltable = new Token[numberColumnsValue];
                for (int j = 0; j < numberColumnsValue; j++) {
                  String totstr = "";
                  Token[] cols = new Token[datavec.length];
                  for (int i = 0; i < datavec.length; i++) {
                    String temp = (String)datavec[i].elementAt(j);
                    StringToken str = new StringToken(temp);
                    cols[i] = str;
                    totstr = totstr + temp + ",";
                  }
                  totstr = totstr.substring(0, totstr.length()-1);
                  StringToken strcol = new StringToken(totstr);
                  strcoltable[j] = strcol;
                  coltable[j] = new ArrayToken(cols);
                }
            output.send(0, new ArrayToken(coltable));
            stringOutput.send(0, new ArrayToken(strcoltable));
            }

        }
    } 


    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

    //The output format the user wishes 
    private int _outputFormat;
    private static final int _ROW = 0;
    private static final int _COLUMN = 1;

}
